//     MagLoopTuner2: Steering program for Magnetic Loop Antenna tuning
//     Schematic and PCB: see MagLoopTuner4.sch and .brd
//
//          (C) PA2EVR 2008               All rights reserved
//
//          PIC16F876A`  Rev: 0.81        7 Dec 2008
//          OSC.......:  XT 8MHz
//
// Structure:
// A Frequency counter is made by Timer1 counting pulses, coming from T1CK1
// pin, preceeded by a 1:16 external prescaler.
// A gate by Tmr2 makes interrupts every 1000 us.
// Interrupts are counted to 16. So, Tmr1 will give the frequency in KHz.
// The freq counter is used to
//   1) Determine the band (in MHz, 1...30), this is also EEPROM band
//   2) make a rough estimate where the Cap should be
//   3) to detect out-of-tune situation
//
// Connections for 16PIC876A:
// Intended for final application of stand-alone tuner
// RA0     = Voltage Reflected
// RA1     = Voltage Forward
// RA2     = Reference voltage (3.3 volt zener)
// RB0     = interrupt for wake up from sleep
// RB1...7 = Leds for program status
// RC0     = Freq in
// RC1...2 = inputs for limit switches
// RC3     = spare input
// RC4...7 = outputs for stepper motor
// RC6     = Serial TX if no stepper connected (test only)

// RS232:  pin 5 = gnd, pin 2 = RX

//==== Constants ===============================================================
#define fcount_time       16         /* gives 16000 micro seconds (exact)     */
#define step_delay         2         /* 2 gives 3 milli seconds (exact)       */
#define settle_delay       6         /* delay before measuring SWR            */
#define max_ratio        163         /* max reflection without retune SWR~2   */
#define max_offdip         5         /* max drift from lowest ratio dip       */
#define max_offtune        3         /* max KHz before retune                 */
#define Search_Width     200         /* number of steps in any band           */

//---- mimic of structure Tuner ------------------------------------------------
#define signal_present    F7
#define freq_valid        F6
#define ratio_valid       F5
#define dip_found         F4
#define tuned             F3

//---- mimic of structure Stepper ----------------------------------------------
#define sp_reached        F7
#define search_dir        F6


//==== Declarations ============================================================
enum            {
                StepEEPROMErase = 0,
                StepVariableReset,
                StepAlignCapMin,     // LED    1: *------
                StepAlignCapMax,     // LED    2: -*-----
                StepStandby,         // LED    3: --*----
                StepDecide,          // 
                StepChangeBand,      // LED    4: ---*---
                StepFineTune,        // LED    5: ----*--
                StepTuned,           // LED    6: -----*-
                StepError            // LED 1..6: ******-
                };                   // LED    7: ......*  indicates signal
                                  
enum            {
                LessCap = 0,
                MoreCap = 1
                };

unsigned int    V_forw,
                V_refl,
                Ratio,
                LowestRatio,
                Raw_kHz,
                Max_kHz,
                Measured_kHz,
                Raw_kHz_Diff,
                Raw_kHz_Prev,
                Last_Tuned_Freq,
                Tuned_Freq_Prev,
                Tuned_Cap_Prev,
                CapPosAct,
                CapPosMax,
                CapPosSetp,
                CapPosSetpMax,
                CapPosSetpMin,
                CapPosActAtDip,
                Wait_Timer,
                LED_Timer,
                Signal_Timer;

unsigned short  ProgramStep,
                Index,
                FcountTmr,
                Measurement_counter,
                StepDelay,
                SettleDelay,
                StepPointer,
                ReversalCounter,
                Measured_band,
                Last_Tuned_band,
                steppattern[4] = {0x20, 0x10, 0x80, 0x40},  // half wave
                Tuner,              // status flags
                Stepper;            // status flags


//---- Forward Declarations ----------------------------------------------------
void            main(void);
void            interrupt(void);
void            DisplayStatusLED(void);
void            EEPROMIndex (unsigned short band);
void            ClearEEPROM(void);
void            WriteTuningToEEPROM(unsigned short band, unsigned freq, unsigned steppos);
unsigned        ReadEEPROM(unsigned short band, unsigned short param);
unsigned        delta_abs(unsigned var1, unsigned var2);
void            wait(unsigned ms);

//==== Functions ===============================================================
void main(void)
   {
   //----- Initialize variables ------------------------------------------------
   ProgramStep          = StepVariableReset;
   Wait_Timer           = 0;
   LED_Timer            = 0;
   Signal_Timer         = 0;
   PORTA                = 0;
   
   //----- Set up Interrupts ---------------------------------------------------
   INTCON.PEIE = 1;               //Enable Peripheral Interrupt (incl TMR1)
   PIE1.TMR2IE = 1;               //Enable Timer 2 interrupt
   INTCON.RBIE = 0;               //Disable PORTB general interrupt
   INTCON.GIE  = 1;               //Enable Gen Interrupt

   //----- Set up Ports --------------------------------------------------------
   TRISA       = 0b00001111;      //RA3..0 ports are inputs
   TRISB       = 0b00000001;      //RB7..1 ports are outputs, RB0 is input
   TRISC       = 0b00001111;      //RC7..4 ports are outputs, RC3..0 are inputs
   OPTION_REG  = 0b11000000;      //Disable weak pull up, INT RB0 on rising edge

   //----- Set up Timer1 as counter for Freq measurement -----------------------
   T1CON       = 0b00000110;      //Tmr1 Presc to 1:1, no osc, no sync,
                                  //input from RC0, stop

   //----- Set up Timer2 as 1000 usec heartbeat generator ----------------------
   //T2CON     = 0b00001101;      //Tmr2 Postsc 1:2, Presc to 1:4, run  // 8 MHz
   T2CON       = 0b00000101;      //Tmr2 Postsc 1:1, Presc to 1:4, run  // 4 MHz
   PR2         = 249;             //Timer 2 period register to 250

   //----- Set up Serial Port (used ony for testing) ---------------------------
   RCSTA.SPEN  = 0;               //disable serial port
   SPBRG       = 12;              //Baudrate = 19200
   TXSTA       = 0b00100100;      //Set bit TXEN and BRGH (high speed)

   //----- Set up ADC ----------------------------------------------------------
   ADCON0      = 0b10000001;      //Fosc/64 and AD on
   ADCON1      = 0b11000101;      //Right justified, Fosc/16, Vref+ = RA3,
                                  //AN0..1 are inputs

   //----- Main Loop -----------------------------------------------------------
   while(1)
      {
      //----- Ratio measurement (similar to SWR) -------------------------------
      // designed for Bruene bridge 5-100W:
      // multiply Vforw by  0.125 (=SHR 3 bits)
      // multiply Vrefl by 64     (=SHL 6 bits)
      // SWR:   1.00 1.01 1.02 1.03 1.05 1.10 1.20 1.50 2.00 3.00
      // Ratio:    0    1    3    4    8   15   35   93  163  250
      V_forw = Adc_Read(1);       //forward power
      V_refl = Adc_Read(0);       //reflected power
      if (V_forw > 25)
         {
         Tuner.ratio_valid = 1;
         Ratio = (V_refl << 6) / (V_forw >> 3);
         }
         else
         {
         Tuner.ratio_valid = 0;
         Ratio = 999;
         }

      //----- Verify if all input signals are present --------------------------
      if (PORTB.F0 && Tuner.freq_valid && Tuner.ratio_valid)
         {
         Tuner.signal_present = 1;
         Signal_Timer = 1000;           // is counted back by interrupt routine
         }
         else
         Tuner.signal_present = 0;

      //----- Force Programstep to StepStandby if no signal --------------------
      if (Signal_Timer == 0 && ProgramStep > StepChangeBand)
         ProgramStep = StepStandby;

      //----- display the steps ------------------------------------------------
      DisplayStatusLED();

      switch (ProgramStep)
         {
         case StepEEPROMErase:
         //----- EEPROM erase, "factory reset" ---------------------------------
            {
            ClearEEPROM(void);
            ProgramStep = StepVariableReset;
            break;
            };

         case StepVariableReset:
         //----- Variable reset ------------------------------------------------
            {
            Tuner                 = 0;       //Clear all flags
            Stepper               = 0;       //Clear all flags
            CapPosSetp            = 0;       // for next step !!
            CapPosAct             = 5000;    // for next step !!
            ProgramStep           = StepAlignCapMin;
            break;
            };

         case StepAlignCapMin:
         //----- Initialize stepper min cap equivalent to max freq -------------
            {
            if (!PORTC.F2 || Stepper.sp_reached)
               {
               CapPosAct          = 0;       // for next step !!
               CapPosSetp         = 5000;    // for next step !!
               Stepper.sp_reached = 0;
               ProgramStep        = StepAlignCapMax;
               }
            break;
            };

         case StepAlignCapMax:
         //----- Initialize stepper max cap equivalent to min freq -------------
            {
            if (!PORTC.F1 || Stepper.sp_reached)
               {
               CapPosMax          = CapPosAct;
               CapPosSetp         = CapPosAct;
               Last_Tuned_Freq    = 0;
               Last_Tuned_band    = 0;
               ProgramStep        = StepStandby;
               }
            break;
            };

         case StepStandby:
         //----- Go to sleep until signal available ----------------------------
            {
            if (!PORTB.F0 && Stepper.sp_reached && Signal_Timer == 0 )
               {
               asm {
                   BCF    INTCON, GIE     //disable global interrupt
                   BCF    INTCON, PEIE    //disable peripherals interrupt
                   BCF    INTCON, INTF    //clear interrupt flag
                   BSF    INTCON, INTE    //enable external interrupt ex RB0
                   SLEEP                  //sleep until pos flank on RB0
                   BCF    INTCON, INTE    //disable external interrupt ex RB0
                   BSF    INTCON, PEIE    //enable peripherals interrupt
                   BSF    INTCON, GIE     //enable global interrupt
                   };
               }
            if (Tuner.signal_present)
               ProgramStep = StepDecide;
            break;
            };

         case StepDecide:
            //----- Is there a band switch? ------------------------------------
            if (Last_Tuned_Band != Measured_Band)
               {
               //set the outermost setpoint limits
               CapPosSetpMin = 20;
               CapPosSetpMax = CapPosMax - 20;
               Tuned_Cap_Prev = CapPosAct;
               Tuned_Freq_Prev = ReadEEPROM(Measured_Band, 0);
               //narrow the setpoint limits if a previous setting exists
               if (Tuned_Freq_Prev > 0 && Tuned_Freq_Prev < 0xFFFF)
                  {
                  Last_Tuned_Freq = Tuned_Freq_Prev;
                  Tuned_Cap_Prev  = ReadEEPROM(Measured_Band, 1);
                  //----- determine Setpoint Limits ----------------------------
                  if (Tuned_Cap_Prev > (Search_Width + CapPosSetpMin))//stay > 0
                     CapPosSetpMin = Tuned_Cap_Prev - Search_Width;
                  if (Tuned_Cap_Prev + Search_Width < CapPosSetpMax)//stay < max
                     CapPosSetpMax = Tuned_Cap_Prev + Search_Width;
                  }
               //----- move to initial setting in this Band --------------------
               CapPosSetp = Tuned_Cap_Prev;
               Stepper.sp_reached = 0;
               Last_Tuned_Band = Measured_Band;
               ProgramStep = StepChangeBand;
               break;
               }

            //----- Re-tuning inside the band? ---------------------------------
            {
            if ((delta_abs(Last_Tuned_Freq, Measured_kHz) > max_offtune) ||
                (Ratio > (LowestRatio + max_offdip))                     ||
                (Ratio > max_ratio))
               {
               Tuner.tuned         = 0;
               ReversalCounter     = 0;
               Tuner.dip_found     = 0;
               LowestRatio         = 999;   //set NOT to 65535 to avoid screw-up
               if (Measured_kHz > Last_Tuned_Freq)
                  Stepper.search_dir = LessCap;
               if (Measured_kHz < Last_Tuned_Freq)
                  Stepper.search_dir = MoreCap;
               ProgramStep = StepFineTune;
               break;
              }

            //----- none of the above, go to the Tuned step --------------------
            ProgramStep = StepTuned;
            break;
            };

         case StepChangeBand:
         //----- Transition of capacitor to estimated position -----------------
         //----- we stay in this step as long as cap is in transit -------------
            {
            if (Stepper.sp_reached || !PORTC.F1 || !PORTC.F2)
               ProgramStep = StepDecide;
            break;
            };

         case StepFineTune:
         //----- Fine tuning based on trial and error --------------------------
            {
            //----- quit if there is no signal ---------------------------------
            if (!Tuner.signal_present) break;

            //----- Start testing for a dip ------------------------------------
            if (Stepper.sp_reached)
               {
               if (Ratio < 300)
                  {
                  if (Ratio < LowestRatio)
                     {
                     LowestRatio = Ratio;
                     CapPosActAtDip = CapPosAct;
                     }
                  if (Ratio > (LowestRatio + max_offdip))
                     {
                     Tuner.dip_found = 1;
                     CapPosSetp = CapPosActAtDip;
                     ProgramStep = StepTuned;
                     break;
                     }
                  }
               if (Stepper.search_dir == MoreCap)      //dip not found, step on
                  {
                  if (CapPosSetp >= CapPosSetpMax || !PORTC.F1)  // High stop
                     {
                     Stepper.search_dir  = LessCap;
                     ReversalCounter++;
                     }
                     else
                     {
                     CapPosSetp++;
                     Stepper.sp_reached = 0;
                     }
                  }
                  else
                  {
                  if (CapPosSetp <= CapPosSetpMin || !PORTC.F2)  // Low stop
                     {
                     Stepper.search_dir  = MoreCap;
                     ReversalCounter++;
                     }
                     else
                     {
                     CapPosSetp--;
                     Stepper.sp_reached = 0;
                     }
                  }
               }
            if (ReversalCounter > 4)
               ProgramStep = StepError;
            break;
            };

         case StepTuned:
         //----- tuned - maintain as long as there is signal -------------------
            {
            if (!Tuner.tuned)
               {
               Last_Tuned_Freq = Measured_kHz;
               Last_Tuned_Band = Measured_Band;
               WriteTuningToEEPROM(Last_Tuned_Band, Last_Tuned_Freq, CapPosAct);
               Tuner.tuned = 1;
               }
            break;
            };

         case StepError:
         //----- error handler -------------------------------------------------
            {
            //Function DisplayStatusLed will show all leds
            WriteTuningToEEPROM(Measured_Band, 0, 0); //erase wrong setting
            LED_Timer   = 2000;
            ProgramStep = StepVariableReset;
            break;
            };

         default :
         //----- catch programming errors --------------------------------------
            {
            ProgramStep = StepVariableReset;
            break;
            };
         }

      }
   }

void interrupt(void)
   {
   //----- Timer 2 provides a 1000 us heartbeat. -------------------------------
   //----- This drives all timing variables ------------------------------------
   if (!PIR1.TMR2IF) return;
   PIR1.TMR2IF  = 0;            // reset interrupt flag ex timer 2
   
   T1CON.TMR1ON = 1;            //Start counting Frequency at new cycle
   //----- if there is no signal, keep freq counter T1 at reset state
   if (!PORTB.F0)
      {
      Max_kHz = 0;              //reset when no signal
      Tuner.freq_valid = 0;     //indicate no valid freq. measurement
      T1CON.TMR1ON = 0;         //counter off
      TMR1H = 0;
      TMR1L = 0;
      FcountTmr = fcount_time ;
      Measurement_counter = 0;
      }
      else
      {
      //----- Frequency Measurement through Tmr1 (configured as counter) -------
      if (FcountTmr == 0)
         {
         T1CON.TMR1ON = 0;         //Stop counting Frequency
         Raw_kHz = ((TMR1H << 8) | TMR1L);
         TMR1H = 0;
         TMR1L = 0;
         FcountTmr = fcount_time ;

         //Check on several measurements, >1 MHz and no overflow
         if (Raw_kHz > 1000 && !PIR1.TMR1IF)
            {
            if (Raw_kHz > Max_kHz)
               Max_kHz = Raw_kHz;     // pick highest count
            if (Raw_kHz > Max_kHz - 3)
               {
               Measurement_counter++;
               if (Measurement_counter >= 2)
                  {
                  Measured_kHz = Max_kHz;
                  Measured_Band = (Max_kHz / 1000);
                  Tuner.freq_valid = 1;
                  Measurement_counter = 2;
                  }
               }
            }
         PIR1.TMR1IF = 0;          //clear Tmr1 overflow flag
         }
      }

   //----- Stepper control, move to setpoint --------------------------------
   if (StepDelay != 0)
      StepDelay-- ;
      else
      {
      StepDelay = step_delay;
      //---Move stepper depending on difference between Setp and ActPos --
      if (CapPosAct == CapPosSetp)
         {
         if (SettleDelay == 0)
            {
            Stepper.sp_reached = 1;
            PORTC = 0;
            }
         }
         else
         {
         Stepper.sp_reached = 0;
         if (CapPosAct < CapPosSetp)
            {
            if (PORTC.F1)
               {
               CapPosAct++;
               StepPointer = (StepPointer + 1) & 0x03;
               }
            }
            else
            {
            if (PORTC.F2)
               {
               CapPosAct--;
               StepPointer = (StepPointer - 1) & 0x03;
               }
            }
         SettleDelay = settle_delay;
         PORTC = steppattern[StepPointer];
         }
      }

   if (FcountTmr    > 0) FcountTmr-- ;    // Frequency counting timer
   if (SettleDelay  > 0) SettleDelay-- ;  // Settle time after step
   if (Wait_Timer   > 0) Wait_Timer-- ;   // Wait Timer
   if (LED_Timer    > 0) LED_Timer-- ;    // Program Timer
   if (Signal_Timer > 0) Signal_Timer-- ; // Signal Timer
   }

void DisplayStatusLED(void)
   {
   switch (ProgramStep)
      {
      // for PA2EVR board
      case StepAlignCapMin   : {PORTB = 0b00000010 ; break;}
      case StepAlignCapMax   : {PORTB = 0b00000100 ; break;}
      case StepStandby       : {PORTB = 0b00100000 ; break;}
      case StepChangeBand    : {PORTB = 0b00010000 ; break;}
      case StepFineTune      : {PORTB = 0b00001000 ; break;}
      case StepTuned         : {PORTB = 0b01000000 ; break;}
      case StepVariableReset : {PORTB = 0b01111110 ; break;}
      default                : {PORTB = 0          ; break;}
      };
//   PORTB = 0;
//   if (PORTB.F0)             PORTB.F4 = 1;
//   if (Tuner.freq_valid)     PORTB.F3 = 1;
//   if (Tuner.ratio_valid)    PORTB.F6 = 1;
   if (Tuner.signal_present) PORTB.F7 = 1;
   while(LED_Timer > 0);      //wait ... ms to show LEDs
   }

void EEPROMIndex (unsigned short Band)
   //for every 1MHz, point to EEPROM address
   {
   if (Band > 52) Band = 52;
   Index = 4 * Band;
   }

void ClearEEPROM(void)
   {
   for (Index = 0; Index < 0xFF; Index++)
      Eeprom_Write(Index,0);
   }
   
void WriteTuningToEEPROM(unsigned short Band, unsigned freq, unsigned steppos)
   //----- write the last known freq and Stepper Position to EEPROM ------------
   //----- Band is per 1 MHz                                      --------------
   {
   EEPROMIndex (Band);
   Eeprom_Write(Index,freq>>8);
   Eeprom_Write(Index+1,freq);
   Eeprom_Write(Index+2,steppos>>8);
   Eeprom_Write(Index+3,steppos);
   wait(20);           //settletime
   }
   
unsigned ReadEEPROM(unsigned short Band, unsigned short param)
   //----- read the last known freq from EEPROM --------------------------------
   {
   EEPROMIndex (Band);
   Index = Index + 2 * param;
   return (Eeprom_Read(Index) << 8) | Eeprom_Read(Index + 1);
   }

unsigned delta_abs(unsigned var1, unsigned var2)
   //----- returns absolute delta between 2 unsigned integers ------------------
   {
   if (var1 > var2)
      return var1 - var2;
      else
      return var2 - var1;
   }

void wait(unsigned ms)
   //----- waiting routine, counter is decreased by interrupt routine ----------
   {
   Wait_Timer = ms;
   while (Wait_Timer > 0);
   }

